/*
 * BK Id: SCCS/s.rdimage.c 1.6 05/18/01 15:17:06 cort
 */
/*
 * rdimage.c
 *
 * Dummy file to allow a compressed initrd to be added
 * into a linker section, accessed by the boot coode
 */

char dummy_for_rdimage;
